#!/usr/bin/env python3

import sys
import torch
import re
#import pdb
import conllu

from transformers import FlaubertModel, FlaubertTokenizer
from argparse import FileType, ArgumentParser
from collections import OrderedDict

parser = ArgumentParser(\
  description="FlauBERT contextual embeddings for FRSemCor nouns")

parser.add_argument("--origfrsemcor", required=True,
    help="ConLLU+ file with FRSemCor sentences (contexts for nouns)",
    type=FileType(mode='r',encoding='utf-8'))
parser.add_argument("--nounsfile", required=True,
    help="TSV with sent ID, noun, ref class + any number of real scores",
    type=FileType(mode='r',encoding='utf-8'))
parser.add_argument("--model", default="small", type=str,
    help="Model $m$, string $m$ \in {small, base, large}")

args = parser.parse_args()
count_incompat = 0

###############################################################################

def find_positions(lemma, tok_text, sentence):
	global count_incompat
	sent_indices = []	
	surfaces = [token["form"]+"</w>" for token in sentence \
				# lemma searched
	            if token['lemma']==lemma and \
	            ":" not in list(token["feats"].keys())[0]]
	len_surf = len(surfaces)
	# special cases for distinct tokenization between BERT and sequoia:
	if 'n°</w>'	in surfaces:
		surfaces.append('°</w>')
	if "fémur</w>" in surfaces:
		surfaces.append('fémur-1,5</w>')
	if "sujets</w>" in surfaces:
		surfaces.append('sujets-</w>')
	if "personnes</w>" in surfaces:
		surfaces.append('personnes-</w>')
	if "président</w>" in surfaces:
		surfaces.append('ex-président</w>')
	if "experts</w>" in surfaces:
		surfaces.append('experts-</w>')
	if "factions</w>" in surfaces:
		surfaces.append('factions-</w>')		
	if "numero(s)</w>" in surfaces:
		surfaces.append('numero</w>')
	if "NUMERO(S)</w>" in surfaces:
		surfaces.append('NUMERO</w>')
	if 'concentration</w>' in surfaces:
		surfaces.append("concentration-dépendante</w>")
	if lemma in ["homme", "femme"]:
		surfaces.append('hommes-femmes</w>')
		surfaces.append('homme-femme</w>')
	if len(surfaces) != len_surf :
		count_incompat += 1
	form = ""
	for i,token in enumerate(tok_text) :		
		if token != "<s>":
			form += token
		if token.endswith("</w>") :
			if form in surfaces :
				sent_indices.append(i)
			form = ""	
	return sent_indices

###############################################################################

FLAUBERT_MODEL = 'flaubert/flaubert_'+args.model+'_cased'

sentences = {}
with args.origfrsemcor as origfrsemcor:
	for tokenlist in conllu.parse_incr(origfrsemcor):
		sentences[tokenlist.metadata["sent_id"]] = tokenlist

# Initialize Flaubert models
flaubert, log = FlaubertModel.from_pretrained(FLAUBERT_MODEL, 
                                              output_loading_info=True)
flaubert_tok = FlaubertTokenizer.from_pretrained(FLAUBERT_MODEL, 
                                                 do_lowercase=False)

with args.nounsfile as nounsfile:
	treated = []
	prev_sentid = ""
	# For each noun in the list of nouns
	for (inoun, nounsline) in enumerate(nounsfile) :		
		fields = nounsline.strip().split("\t")
		if prev_sentid != fields[0]: # The noun is in a new sentence
			treated = []
		sentence = sentences[fields[0]] # get sentence for sent_id
		nounlemma = fields[1]
		ref = fields[2]
		print("Processing noun number {}: {}".format(inoun + 1, nounlemma),
		       file=sys.stderr)
		text = sentence.metadata["text"]
		tok_ids = flaubert_tok.encode(text)
		tok_text = flaubert_tok.convert_ids_to_tokens(tok_ids)
		token_ids = torch.tensor([tok_ids])
		indices = find_positions(nounlemma,tok_text,sentence)		
		if len(indices) == 1 : # the noun appears once in the sentence
			embedding = flaubert(token_ids)[0][:, indices[0], :]			
		else :	# the noun appears more than once in the sentence
			try:	
				embedding = flaubert(token_ids)[0][:, indices[len(treated)], :]
				# keep track of treated occurrences of noun in current sentence
				treated.append(positions[len(treated)]) 	
			except Exception: # ugly workaround for tokenisation incompatibilities								
				count_incompat += 1
				# Arbitrarily take the embedding of the first occurrence				
				embedding = flaubert(token_ids)[0][:, indices[0], :]				
		strembedding = "\t".join(map(lambda x: str(x), embedding.tolist()[0]))
		print("\t".join(fields[:3]) + "\t" + strembedding)
		prev_sentid = fields[0]

print("Tokenisation incompatibilities: {}".format(count_incompat),
      file=sys.stderr)
# test: 50/1013
# dev: 31/1015
# train: 574/5160
